@@ -19,6 +19,7 @@ import collections
 import copy
 import json
 import os
+import re
 import sys
 import warnings
 
@@ -68,6 +69,8 @@ VENDOR_FILES = [
 
 BOOL_KEYS = ('insecure', 'cache')
 
+FORMAT_EXCLUSIONS = frozenset(['password'])
+
 
 # NOTE(dtroyer): This turns out to be not the best idea so let's move
 #                overriding defaults to a kwarg to OpenStackConfig.__init__()
@@ -105,9 +108,12 @@ def _get_os_environ(envvar_prefix=None):
     for k in environkeys:
         newkey = k.split('_', 1)[-1].lower()
         ret[newkey] = os.environ[k]
-    # If the only environ keys are cloud and region_name, don't return anything
-    # because they are cloud selectors
-    if set(environkeys) - set(['OS_CLOUD', 'OS_REGION_NAME']):
+    # If the only environ keys are selectors or behavior modification, don't
+    # return anything
+    selectors = set([
+        'OS_CLOUD', 'OS_REGION_NAME',
+        'OS_CLIENT_CONFIG_FILE', 'OS_CLIENT_SECURE_FILE', 'OS_CLOUD_NAME'])
+    if set(environkeys) - selectors:
         return ret
     return None
 
@@ -146,7 +152,9 @@ def _fix_argv(argv):
     # over the place.
     processed = collections.defaultdict(list)
     for index in range(0, len(argv)):
-        if argv[index].startswith('--'):
+        # If the value starts with '--' and has '-' or '_' in it, then
+        # it's worth looking at it
+        if re.match('^--.*(_|-)+.*', argv[index]):
             split_args = argv[index].split('=')
             orig = split_args[0]
             new = orig.replace('_', '-')
@@ -173,19 +181,28 @@ class OpenStackConfig(object):
     def __init__(self, config_files=None, vendor_files=None,
                  override_defaults=None, force_ipv4=None,
                  envvar_prefix=None, secure_files=None,
-                 pw_func=None, session_constructor=None):
+                 pw_func=None, session_constructor=None,
+                 app_name=None, app_version=None,
+                 load_yaml_config=True):
         self.log = _log.setup_logging(__name__)
         self._session_constructor = session_constructor
+        self._app_name = app_name
+        self._app_version = app_version
 
-        self._config_files = config_files or CONFIG_FILES
-        self._secure_files = secure_files or SECURE_FILES
-        self._vendor_files = vendor_files or VENDOR_FILES
+        if load_yaml_config:
+            self._config_files = config_files or CONFIG_FILES
+            self._secure_files = secure_files or SECURE_FILES
+            self._vendor_files = vendor_files or VENDOR_FILES
+        else:
+            self._config_files = []
+            self._secure_files = []
+            self._vendor_files = []
 
-        config_file_override = os.environ.pop('OS_CLIENT_CONFIG_FILE', None)
+        config_file_override = os.environ.get('OS_CLIENT_CONFIG_FILE')
         if config_file_override:
             self._config_files.insert(0, config_file_override)
 
-        secure_file_override = os.environ.pop('OS_CLIENT_SECURE_FILE', None)
+        secure_file_override = os.environ.get('OS_CLIENT_SECURE_FILE')
         if secure_file_override:
             self._secure_files.insert(0, secure_file_override)
 
@@ -214,12 +231,12 @@ class OpenStackConfig(object):
         else:
             # Get the backwards compat value
             prefer_ipv6 = get_boolean(
-                os.environ.pop(
+                os.environ.get(
                     'OS_PREFER_IPV6', client_config.get(
                         'prefer_ipv6', client_config.get(
                             'prefer-ipv6', True))))
             force_ipv4 = get_boolean(
-                os.environ.pop(
+                os.environ.get(
                     'OS_FORCE_IPV4', client_config.get(
                         'force_ipv4', client_config.get(
                             'broken-ipv6', False))))
@@ -231,7 +248,7 @@ class OpenStackConfig(object):
                 self.force_ipv4 = True
 
         # Next, process environment variables and add them to the mix
-        self.envvar_key = os.environ.pop('OS_CLOUD_NAME', 'envvars')
+        self.envvar_key = os.environ.get('OS_CLOUD_NAME', 'envvars')
         if self.envvar_key in self.cloud_config['clouds']:
             raise exceptions.OpenStackConfigException(
                 '"{0}" defines a cloud named "{1}", but'
@@ -239,9 +256,8 @@ class OpenStackConfig(object):
                 ' either your environment based cloud, or one of your'
                 ' file-based clouds.'.format(self.config_filename,
                                              self.envvar_key))
-        # Pull out OS_CLOUD so that if it's the only thing set, do not
-        # make an envvars cloud
-        self.default_cloud = os.environ.pop('OS_CLOUD', None)
+
+        self.default_cloud = os.environ.get('OS_CLOUD')
 
         envvars = _get_os_environ(envvar_prefix=envvar_prefix)
         if envvars:
@@ -249,6 +265,17 @@ class OpenStackConfig(object):
             if not self.default_cloud:
                 self.default_cloud = self.envvar_key
 
+        if not self.default_cloud and self.cloud_config['clouds']:
+            if len(self.cloud_config['clouds'].keys()) == 1:
+                # If there is only one cloud just use it. This matches envvars
+                # behavior and allows for much less typing.
+                # TODO(mordred) allow someone to mark a cloud as "default" in
+                # clouds.yaml.
+                # The next/iter thing is for python3 compat where dict.keys
+                # returns an iterator but in python2 it's a list.
+                self.default_cloud = next(iter(
+                    self.cloud_config['clouds'].keys()))
+
         # Finally, fall through and make a cloud that starts with defaults
         # because we need somewhere to put arguments, and there are neither
         # config files or env vars
@@ -525,6 +552,7 @@ class OpenStackConfig(object):
             network = dict(
                 name=name,
                 routes_externally=get_boolean(net.get('routes_externally')),
+                nat_source=get_boolean(net.get('nat_source')),
                 nat_destination=get_boolean(net.get('nat_destination')),
                 default_interface=get_boolean(net.get('default_interface')),
             )
@@ -609,19 +637,15 @@ class OpenStackConfig(object):
                 'tenant_id', 'tenant-id', 'project_id', 'project-id')
             mappings['project_name'] = (
                 'tenant_name', 'tenant-name', 'project_name', 'project-name')
-        # Special-case username and password so that we don't have to load
-        # the plugins so early
-        mappings['username'] = ('user-name', 'user_name', 'username')
-        mappings['password'] = ('password',)
         for target_key, possible_values in mappings.items():
             target = None
             for key in possible_values:
-                root_target = cloud.pop(key, None)
-                auth_target = cloud['auth'].pop(key, None)
-                if root_target:
-                    target = str(root_target)
-                elif auth_target:
-                    target = str(auth_target)
+                if key in cloud:
+                    target = str(cloud[key])
+                    del cloud[key]
+                if key in cloud['auth']:
+                    target = str(cloud['auth'][key])
+                    del cloud['auth'][key]
             if target:
                 cloud['auth'][target_key] = target
         return cloud
@@ -651,6 +675,7 @@ class OpenStackConfig(object):
         the keystoneauth Auth Plugin Options and os-cloud. Also, peek in the
         argv to see if all of the auth plugin options should be registered
         or merely the ones already configured.
+
         :param argparse.ArgumentParser: parser to attach argparse options to
         :param list argv: the arguments provided to the application
         :param string service_keys: Service or list of services this argparse
@@ -847,17 +872,16 @@ class OpenStackConfig(object):
         return config
 
     def _get_auth_loader(self, config):
-        # Re-use the admin_token plugin for the "None" plugin
+        # Use the 'none' plugin for variants of None specified,
         # since it does not look up endpoints or tokens but rather
         # does a passthrough. This is useful for things like Ironic
         # that have a keystoneless operational mode, but means we're
         # still dealing with a keystoneauth Session object, so all the
         # _other_ things (SSL arg handling, timeout) all work consistently
         if config['auth_type'] in (None, "None", ''):
-            config['auth_type'] = 'admin_token'
-            # Set to notused rather than None because validate_auth will
-            # strip the value if it's actually python None
-            config['auth']['token'] = 'notused'
+            # 'none' auth_plugin has token as 'notused' so validate_auth will
+            # not strip the value (which it does for actual python None)
+            config['auth_type'] = 'none'
         elif config['auth_type'] == 'token_endpoint':
             # Humans have been trained to use a thing called token_endpoint
             # That it does not exist in keystoneauth is irrelvant- it not
@@ -939,7 +963,7 @@ class OpenStackConfig(object):
     def option_prompt(self, config, p_opt):
         """Prompt user for option that requires a value"""
         if (
-                p_opt.prompt is not None and
+                getattr(p_opt, 'prompt', None) is not None and
                 p_opt.dest not in config['auth'] and
                 self._pw_callback is not None
         ):
@@ -972,7 +996,6 @@ class OpenStackConfig(object):
                 ('auth_token' in config and config['auth_token']) or
                 ('token' in config and config['token'])):
             config.setdefault('token', config.pop('auth_token', None))
-            config.setdefault('auth_type', 'token')
 
         # These backwards compat values are only set via argparse. If it's
         # there, it's because it was passed in explicitly, and should win
@@ -1019,6 +1042,7 @@ class OpenStackConfig(object):
         :raises: keystoneauth1.exceptions.MissingRequiredOptions
             on missing required auth parameters
         """
+
         args = self._fix_args(kwargs, argparse=argparse)
 
         if cloud is None:
@@ -1058,27 +1082,15 @@ class OpenStackConfig(object):
         config = self.auth_config_hook(config)
 
         if validate:
-            try:
-                loader = self._get_auth_loader(config)
-                config = self._validate_auth(config, loader)
-                auth_plugin = loader.load_from_options(**config['auth'])
-            except Exception as e:
-                # We WANT the ksa exception normally
-                # but OSC can't handle it right now, so we try deferring
-                # to ksc. If that ALSO fails, it means there is likely
-                # a deeper issue, so we assume the ksa error was correct
-                self.log.debug("Deferring keystone exception: {e}".format(e=e))
-                auth_plugin = None
-                try:
-                    config = self._validate_auth_ksc(config, cloud)
-                except Exception:
-                    raise e
+            loader = self._get_auth_loader(config)
+            config = self._validate_auth(config, loader)
+            auth_plugin = loader.load_from_options(**config['auth'])
         else:
             auth_plugin = None
 
         # If any of the defaults reference other values, we need to expand
         for (key, value) in config.items():
-            if hasattr(value, 'format'):
+            if hasattr(value, 'format') and key not in FORMAT_EXCLUSIONS:
                 config[key] = value.format(**config)
 
         force_ipv4 = config.pop('force_ipv4', self.force_ipv4)
@@ -1098,6 +1110,8 @@ class OpenStackConfig(object):
             auth_plugin=auth_plugin,
             openstack_config=self,
             session_constructor=self._session_constructor,
+            app_name=self._app_name,
+            app_version=self._app_version,
         )
 
     def get_one_cloud_osc(
@@ -1171,7 +1185,7 @@ class OpenStackConfig(object):
 
         # If any of the defaults reference other values, we need to expand
         for (key, value) in config.items():
-            if hasattr(value, 'format'):
+            if hasattr(value, 'format') and key not in FORMAT_EXCLUSIONS:
                 config[key] = value.format(**config)
 
         force_ipv4 = config.pop('force_ipv4', self.force_ipv4)
@@ -1223,6 +1237,7 @@ class OpenStackConfig(object):
         with open(config_file, 'w') as fh:
             yaml.safe_dump(cur_config, fh, default_flow_style=False)
 
+
 if __name__ == '__main__':
     config = OpenStackConfig().get_all_clouds()
     for cloud in config:
